﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Runtime.InteropServices;
using System.Collections;

using com.ipevo.Presenter.CameraKit;
using com.ipevo.Presenter.ToolKit;

namespace com.ipevo.Presenter.demo
{
    public partial class MainWindow : Window
    {
        CameraManager cameraManager;
        private Camera previousSelectedCamera = null;
        private StreamProxy.StreamObserverEX cb_streamObserverEX;
        WriteableBitmap writeableBitmap;
        List<VideoResolutionFormat> cameraSupportVideoFormat;

        int source_width = 800;
        int source_height = 600;


        [DllImport("kernel32.dll", SetLastError = true)]
        static extern IntPtr CreateFileMapping(IntPtr hFile, IntPtr lpFileMappingAttributes, uint flProtect, uint dwMaximumSizeHigh, uint dwMaximumSizeLow, string lpName);

        [DllImport("kernel32.dll", SetLastError = true)]
        static extern IntPtr MapViewOfFile(IntPtr hFileMappingObject, uint dwDesiredAccess, uint dwFileOffsetHigh, uint dwFileOffsetLow, uint dwNumberOfBytesToMap);

        [DllImport("Kernel32.dll", EntryPoint = "RtlMoveMemory")]
        private static extern void CopyMemory(IntPtr Destination, IntPtr Source, int Length);

        public MainWindow()
        {
            InitializeComponent();

            NotificationCenter.defaultCenter.addNotificationObserver(CameraManager.Notification.DeviceDetached, new NotificationCenter.NotificationObserver(update_CameraList));
            NotificationCenter.defaultCenter.addNotificationObserver(CameraManager.Notification.DeviceAttached, new NotificationCenter.NotificationObserver(update_CameraList));

            NotificationCenter.defaultCenter.addNotificationObserver(CameraManager.Notification.HID_SNAP_SHOT_BTN , new NotificationCenter.NotificationObserver(camera_snap_shot_click));
            NotificationCenter.defaultCenter.addNotificationObserver(CameraManager.Notification.HID_FOCUS_BEGIN, new NotificationCenter.NotificationObserver(camera_focus_status));
            NotificationCenter.defaultCenter.addNotificationObserver(CameraManager.Notification.HID_FOCUS_FINISH, new NotificationCenter.NotificationObserver(camera_focus_status));
            NotificationCenter.defaultCenter.addNotificationObserver(CameraManager.Notification.HID_S_AF, new NotificationCenter.NotificationObserver(camera_focus_mode));
            NotificationCenter.defaultCenter.addNotificationObserver(CameraManager.Notification.HID_C_AF, new NotificationCenter.NotificationObserver(camera_focus_mode));

            NotificationCenter.defaultCenter.addNotificationObserver(CameraManager.Notification.CAMERA_EXPOSURE_CHANGE, new NotificationCenter.NotificationObserver(camera_exposure_value));            


            cameraManager = CameraManager.sharedManager;
            #region Camera List
            cameraManager.startMonitor();
            this.cameraListComboBox.ItemsSource = cameraManager.cameras;            

            
            #endregion

            cb_streamObserverEX = new StreamProxy.StreamObserverEX(this.update_ucCameraModeImageEX);
            
            writeableBitmap = new WriteableBitmap(source_width, source_height, 96, 96, PixelFormats.Bgr32, null);
            cameraImage.Source = writeableBitmap;

            appVer.Content = "v" + System.Diagnostics.FileVersionInfo.GetVersionInfo(System.Reflection.Assembly.GetExecutingAssembly().Location).FileVersion.ToString();
            this.Closed += MainWindow_Closed;
        }

        void MainWindow_Closed(object sender, EventArgs e)
        {
            NotificationCenter.defaultCenter.removeNotificationObserver(CameraManager.Notification.DeviceDetached, new NotificationCenter.NotificationObserver(update_CameraList));
            NotificationCenter.defaultCenter.removeNotificationObserver(CameraManager.Notification.DeviceAttached, new NotificationCenter.NotificationObserver(update_CameraList));

            NotificationCenter.defaultCenter.removeNotificationObserver(CameraManager.Notification.HID_SNAP_SHOT_BTN, new NotificationCenter.NotificationObserver(camera_snap_shot_click));
            NotificationCenter.defaultCenter.removeNotificationObserver(CameraManager.Notification.HID_FOCUS_BEGIN, new NotificationCenter.NotificationObserver(camera_focus_status));
            NotificationCenter.defaultCenter.removeNotificationObserver(CameraManager.Notification.HID_FOCUS_FINISH, new NotificationCenter.NotificationObserver(camera_focus_status));
            NotificationCenter.defaultCenter.removeNotificationObserver(CameraManager.Notification.HID_S_AF, new NotificationCenter.NotificationObserver(camera_focus_mode));
            NotificationCenter.defaultCenter.removeNotificationObserver(CameraManager.Notification.HID_C_AF, new NotificationCenter.NotificationObserver(camera_focus_mode));

            NotificationCenter.defaultCenter.removeNotificationObserver(CameraManager.Notification.CAMERA_EXPOSURE_CHANGE, new NotificationCenter.NotificationObserver(camera_exposure_value));            

            cameraManager.stopMonitor();        //"must" call this before your Application Exit
        }        

        private void update_CameraList(string str, object sender, object camera)
        {
            updateCameraList();
        }

        private void camera_snap_shot_click(string str, object sender, object userInfo)
        {
            //parameter1 sender
            //parameter2 userInfo is Hashtable
            // userInfo["Camera"]  object of Class Camera
            Console.WriteLine((( (userInfo as Hashtable)["Camera"] ) as Camera).CameraInstanceName + " snap shot button is clicked.");

            if (!this.Dispatcher.CheckAccess())
            {
                this.Dispatcher.Invoke(new dele_SaveCameraImage(SaveCameraImage));
            }
            else
            {
                if ( (userInfo as Hashtable)["Camera"] as Camera == (this.cameraListComboBox.SelectedItem as Camera) )
                {                            
                    SaveCameraImage();
                }
            }
        }

        private void camera_focus_status(string str, object sender, object userInfo)
        {
            //parameter1, sender from
            //parameter2, userInfo is hashtable
            // userInfo["Camera"]: type Camera 
            // userInfo["FocusStatus"]: type int, 0 begin, 1 finish
                        
            if( (int)((userInfo as Hashtable)["FocusStatus"]) == 0 )
            {
                Console.WriteLine(((userInfo as Hashtable)["Camera"] as Camera).CameraInstanceName.ToString() + " " +
                                    "Focus Start");
                change_msg_text( ((userInfo as Hashtable)["Camera"] as Camera).CameraInstanceName.ToString() + " " + "Focus Start" );
            }
            else if( (int)((userInfo as Hashtable)["FocusStatus"]) == 1 )
            {
                Console.WriteLine(((userInfo as Hashtable)["Camera"] as Camera).CameraInstanceName.ToString() + " " +
                                    "Focus Finish");
                change_msg_text( ((userInfo as Hashtable)["Camera"] as Camera).CameraInstanceName.ToString() + " " + "Focus Finish");
            }
        }        

        private void camera_focus_mode(string str, object sender, object userInfo)
        {
            //parameter1, sender from
            //parameter2, userInfo is hashtable
            // userInfo["Camera"]: type Camera 
            // userInfo["FocusMode"]: type int, 0 S-AF, 1 C-AF

            if ((int)((userInfo as Hashtable)["FocusMode"]) == 0)
            {
                Console.WriteLine(((userInfo as Hashtable)["Camera"] as Camera).CameraInstanceName.ToString() + " " +
                                    "S-AF");
                change_msg_text( ((userInfo as Hashtable)["Camera"] as Camera).CameraInstanceName.ToString() + " " + "S-AF");
            }
            else if ((int)((userInfo as Hashtable)["FocusMode"]) == 1)
            {
                Console.WriteLine(((userInfo as Hashtable)["Camera"] as Camera).CameraInstanceName.ToString() + " " +
                                    "C-AF");
                change_msg_text( ((userInfo as Hashtable)["Camera"] as Camera).CameraInstanceName.ToString() + " " + "C-AF");
            }
        }


        private delegate void dele_func(string str, object sender, object userInfo);
        private void camera_exposure_value(string str, object sender, object userInfo)
        {
            //parameter1, sender from
            //parameter2, userInfo is hashtable
            // userInfo["Camera"]: type Camera
            // userInfo["Lock"]: type bool
            // userInfo["ExposureValue"]: type int              

            if (!this.Dispatcher.CheckAccess())
            {
                this.Dispatcher.Invoke(new dele_func(camera_exposure_value), str, sender, userInfo);
            }
            else
            {
                System.Collections.Hashtable exposureInformation = userInfo as System.Collections.Hashtable;
                Console.WriteLine((exposureInformation["Camera"] as Camera).CameraInstanceName +
                                   " EV " + ((int)exposureInformation["ExposureValue"]) +
                                   " AutoExposure Lock is " + ((bool)exposureInformation["Lock"]));

                if ((exposureInformation["Camera"] as Camera) == (this.cameraListComboBox.SelectedItem as Camera))
                {
                    autoExposureSubmenu.Value = ((int)exposureInformation["ExposureValue"]);
                    AE_Lock.IsChecked = ((bool)exposureInformation["Lock"]);
                }
            }
        }

        


        delegate void delegateUpdateCameraList();
        private void updateCameraList()
        {
            if (!this.Dispatcher.CheckAccess())
            {
                this.Dispatcher.Invoke(new delegateUpdateCameraList(updateCameraList));
            }
            else
            {                
                this.cameraListComboBox.ItemsSource = CameraManager.sharedManager.cameras;                                    
            }
        }

        private void cameraList_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {            
            Camera selectedCamera = (sender as ComboBox).SelectedItem as Camera;
            
            if (this.previousSelectedCamera != null)
                StreamProxy.sharedProxy.removeStreamObserverEX(this.previousSelectedCamera, cb_streamObserverEX);
                        
            if (selectedCamera != null)
            {
                cameraSupportVideoFormat = selectedCamera.supportedFormats();

                this.resolutionList.ItemsSource = cameraSupportVideoFormat;
                this.resolutionList.SelectedIndex = 0;                
                //setCameraResolution();            

                if (selectedCamera.hasCapability(Camera.Capability.WhiteBalance) == true)
                {
                    this.autoWhitebalanceSubmenu.Visibility = System.Windows.Visibility.Visible;
                }
                else
                {
                    this.autoWhitebalanceSubmenu.Visibility = System.Windows.Visibility.Hidden;
                }
            }            

            if (selectedCamera != null)
            {
                StreamProxy.sharedProxy.addStreamObserverEX(selectedCamera, cb_streamObserverEX);
                
                string firmwareVer = "";
                selectedCamera.getFirmwareVersion(out firmwareVer);
                this.firmwareVerLabel.Content = (firmwareVer);

                //get UVC camera default settings
                AdjustCamera_GetCurrent(Camera.PropertyValueType.Default);              
            }
            previousSelectedCamera = selectedCamera;           
        }

        private void buttonFocus_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                Camera selectedCamera = cameraListComboBox.SelectedItem as Camera;
                selectedCamera.setFocus(false, 0);
                selectedCamera.startFocus();
            }
            catch (Exception)
            { }
        }


        private void update_ucCameraModeImageEX(Camera camera, IntPtr pBitmap, int size, DirectShowLib.IMediaSample mediaSample)
        {            
            updateImageContent(camera, pBitmap, size);
            return;
        }
        
        delegate void deleUpdateImageContent(Camera camera, IntPtr pBmp, int bufferLength);
        private void updateImageContent(Camera camera, IntPtr pBmp, int bufferLength)
        {
            if (!this.Dispatcher.CheckAccess())
            {               
                this.Dispatcher.Invoke(new deleUpdateImageContent(updateImageContent), camera, pBmp, bufferLength);
            }
            else
            {               
                if (writeableBitmap != null && pBmp != IntPtr.Zero)
                {
                    writeableBitmap.Lock();
                    Int32Rect rect = new Int32Rect(0, 0, (int)writeableBitmap.Width, (int)writeableBitmap.Height);
                    IntPtr backbuffer = writeableBitmap.BackBuffer;
                    CopyMemory(backbuffer, pBmp, bufferLength);
                    writeableBitmap.AddDirtyRect(rect);
                    writeableBitmap.Unlock();                                    
                }
            }
        }

        private delegate void deleSetCameraResolution();
        private void setCameraResolution()
        {            
            if (!this.Dispatcher.CheckAccess())
            {               
                this.Dispatcher.Invoke(new deleSetCameraResolution(setCameraResolution));
            }
            else
            {                
                Camera cam = (cameraListComboBox.SelectedItem as Camera);

                if (this.resolutionList.SelectedIndex >= 0)
                {
                    cam.setFormat(cameraSupportVideoFormat[this.resolutionList.SelectedIndex]);

                    StreamProxy.sharedProxy.stopStreamObserver(cam);

                    if (writeableBitmap != null)
                        writeableBitmap = null;

                    source_width = cameraSupportVideoFormat[this.resolutionList.SelectedIndex].width;
                    source_height = cameraSupportVideoFormat[this.resolutionList.SelectedIndex].height;
                    writeableBitmap = new WriteableBitmap(source_width, source_height, 96, 96, PixelFormats.Bgr32, null);
                    cameraImage.Source = writeableBitmap;

                    StreamProxy.sharedProxy.startStreamObserver(cam);
                }
            }
        }

        private void resolutionList_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            setCameraResolution();
        }

        private void autoExposureSubmenu_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            try
            {
                Camera selectedCamera = cameraListComboBox.SelectedItem as Camera;
                selectedCamera.setExposure((short)(e.NewValue));
            }
            catch(Exception)
            { }
        }

        private void autoWhitebalanceSubmenu_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            try
            {
                Camera selectedCamera = cameraListComboBox.SelectedItem as Camera;
                selectedCamera.setWhiteBalance( (short)(e.NewValue) );
            }
            catch (Exception)
            { }
        }

        private void manualFocusSubmenu_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            try
            {
                Camera selectedCamera = cameraListComboBox.SelectedItem as Camera;
                selectedCamera.setFocus( true, (short)(e.NewValue));
            }
            catch (Exception)
            { }
        }

        private void AdjustCamera(string adjType, double adjValue)
        {
            try
            {                
                if (adjType == "brightness")
                {                
                    CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].setBrightness((short)adjValue);                
                }
                else if (adjType == "contrast")
                {
                    CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].setContrast((short)adjValue);                
                }
                else if (adjType == "gamma")
                {
                    CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].setGamma((short)adjValue);                
                }
                else if (adjType == "hue")
                {
                    CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].setHue((short)adjValue);                
                }
                else if (adjType == "saturation")
                {
                    CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].setSaturation((short)adjValue);                
                }
                else if (adjType == "sharpness")
                {
                    CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].setSharpness((short)adjValue);                
                }
            }
            catch (Exception)
            { } 
        }

        private void brightnessSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            AdjustCamera("brightness", e.NewValue);                      
        }

        private void contrastSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            AdjustCamera("contrast", e.NewValue);
        }

        private void gammaSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            AdjustCamera("gamma", e.NewValue);
        }

        private void hueSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            AdjustCamera("hue", e.NewValue);
        }

        private void saturationSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            AdjustCamera("saturation", e.NewValue);
        }

        private void sharpnessSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            AdjustCamera("sharpness", e.NewValue);
        }

        private void AdjustCamera_GetCurrent(Camera.PropertyValueType wantType)
        {
            short getValue, max, min;
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getBrightness(out getValue, wantType);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getBrightness(out min, Camera.PropertyValueType.Minimum);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getBrightness(out max, Camera.PropertyValueType.Maximum);
            this.brightnessSlider.Maximum = max;
            this.brightnessSlider.Minimum = min;
            if (this.brightnessSlider.IsEnabled)
                this.brightnessSlider.Value = getValue;

            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getContrast(out getValue, wantType);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getContrast(out min, Camera.PropertyValueType.Minimum);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getContrast(out max, Camera.PropertyValueType.Maximum);
            this.contrastSlider.Maximum = max;
            this.contrastSlider.Minimum = min;
            this.contrastSlider.Value = getValue;

            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getGamma(out getValue, wantType);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getGamma(out min, Camera.PropertyValueType.Minimum);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getGamma(out max, Camera.PropertyValueType.Maximum);
            this.gammaSlider.Maximum = max;
            this.gammaSlider.Minimum = min;
            this.gammaSlider.Value = getValue;

            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getHue(out getValue, wantType);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getHue(out min, Camera.PropertyValueType.Minimum);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getHue(out max, Camera.PropertyValueType.Maximum);
            this.hueSlider.Maximum = max;
            this.hueSlider.Minimum = min;
            this.hueSlider.Value = getValue;

            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getSaturation(out getValue, wantType);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getSaturation(out min, Camera.PropertyValueType.Minimum);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getSaturation(out max, Camera.PropertyValueType.Maximum);
            this.saturationSlider.Maximum = max;
            this.saturationSlider.Minimum = min;
            this.saturationSlider.Value = getValue;

            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getSharpness(out getValue, wantType);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getSharpness(out min, Camera.PropertyValueType.Minimum);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getSharpness(out max, Camera.PropertyValueType.Maximum);
            this.sharpnessSlider.Maximum = max;
            this.sharpnessSlider.Minimum = min;
            this.sharpnessSlider.Value = getValue;

            short focus_range_min;
            short focus_range_max;
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getFocusValue(out focus_range_min, Camera.PropertyValueType.Minimum);
            CameraManager.sharedManager.cameras[this.cameraListComboBox.SelectedIndex].getFocusValue(out focus_range_max, Camera.PropertyValueType.Maximum);
            this.manualFocusSubmenu.Maximum = focus_range_max;
            this.manualFocusSubmenu.Minimum = focus_range_min;            
        }
      
        private void reset_uvc_adjustment_Click(object sender, RoutedEventArgs e)
        {
            AdjustCamera_GetCurrent(Camera.PropertyValueType.Default);
        }
        
        private void buttonSnapShot_Click(object sender, RoutedEventArgs e)
        {
            SaveCameraImage();            
        }

        private delegate void dele_SaveCameraImage();
        private void SaveCameraImage()
        {
            System.Drawing.Bitmap snapBmp = BitmapSourceToBmp((BitmapSource)writeableBitmap);
            string fileName = AppDomain.CurrentDomain.BaseDirectory + DateTime.Now.ToString("MM-dd-yyyy_HHmmss") + ".jpg";
            snapBmp.Save(fileName);

            //open current folder
            System.Diagnostics.Process.Start(AppDomain.CurrentDomain.BaseDirectory);
        }

        public System.Drawing.Bitmap BitmapSourceToBmp(BitmapSource bitmapsource)
        {

            System.Drawing.Bitmap bmp = new System.Drawing.Bitmap(bitmapsource.PixelWidth, bitmapsource.PixelHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb);
            System.Drawing.Imaging.BitmapData data = bmp.LockBits(
                new System.Drawing.Rectangle(System.Drawing.Point.Empty, bmp.Size),
                System.Drawing.Imaging.ImageLockMode.WriteOnly,
                System.Drawing.Imaging.PixelFormat.Format32bppRgb);

            bitmapsource.CopyPixels(
                System.Windows.Int32Rect.Empty,
                data.Scan0,
                data.Height * data.Stride,
                data.Stride);
            bmp.UnlockBits(data);
            return bmp;       
        }

        private void AE_Lock_Change(object sender, RoutedEventArgs e)
        {
            try
            {
                Camera selectedCamera = cameraListComboBox.SelectedItem as Camera;
                selectedCamera.setAutoExposureLock((bool)(AE_Lock.IsChecked));
            }
            catch( Exception ){}
        }

        public delegate void dele_change_msg_text(string str);
        public void change_msg_text(string str)
        {
            if (!this.Dispatcher.CheckAccess())
            {
                this.Dispatcher.Invoke(new dele_change_msg_text(change_msg_text), str);
            }
            else
            {
                msg_text.Text = str;
            }
        }
    }
}
